# frozen_string_literal: true

require 'spec_helper'

RSpec.describe Heptapod::NativeMercurial::Migrator do
  let(:project) { create(:hg_git_project, :hg_repository)}
  let(:repo) { project.repository }
  let(:user) { build(:user) }
  let(:prepare_repo) do
    next if repo.nil? # some tests don't use a repo

    repo.hg_call(%w[up 0])
    hg_commit(1)
    repo.hg_call(%w[up 0])
    hg_commit(2, topic: 'feature')
  end

  let(:hg_sha) { repo.hgsha_from_rev('feature') }
  let(:git_sha) { repo.sha_from_hgsha(hg_sha) }
  let(:hg_base_sha) { repo.hgsha_from_rev('default') }
  let(:git_base_sha) { repo.sha_from_hgsha(hg_base_sha) }

  subject do
    sub = described_class.new(project, project.creator)
    sub.init_sha_map
    sub
  end

  def hg_commit(cid, topic: nil)
    if topic
      repo.hg_call(['topic', topic])
    end

    repo.write_file(repo.hg_full_path + "/file#{cid}", "line\n")
    repo.hg_call(["add", "file#{cid}"])
    repo.hg_call(['ci', '-m', "Commit #{cid}", '--config',
                  'experimental.single-head-per-branch=False'],
                 user: user, for_write: true)
    repo.hg_git_invalidate_maps!
  end

  before do
    prepare_repo
  end

  context 'migration progress and exclusive lease management' do
    let(:project) { create(:hg_git_project) }
    let(:repo) { nil }
    let(:subject) { described_class.new(project, project.creator) }

    describe '#enter' do
      it "does not crash and executes the inner block" do
        # catch basic refactoring mistakes
        x = 1
        subject.enter { x += 1 }
        expect(x).to eq(2)
      end
    end

    describe "#try_obtain_exclusive_lease" do
      it "acquires and releases the lease properly in normal operation" do
        subject.try_obtain_exclusive_lease do
          expect(subject.exclusive_lease_taken?).to be true
        end
        expect(subject.exclusive_lease_taken?).to be false
      end

      it "releases the lease properly in case of error and does not swallow exceptions" do
        internal_error_msg = "oops, that was unexpected"

        expect do
          subject.try_obtain_exclusive_lease do
            raise StandardError, internal_error_msg
          end
        end.to raise_error.with_message(internal_error_msg)

        expect(subject.exclusive_lease_taken?).to be false
      end

      it "skips the block if lease is already taken" do
        subject.try_obtain_exclusive_lease do
          expect(subject.exclusive_lease_taken?).to be true
          subject.try_obtain_exclusive_lease do
            raise StandardError, "This block should not be executed"
          end
        end
        expect(subject.exclusive_lease_taken?).to be false
      end
    end

    describe "#renew_exclusive_lease" do
      it "does not crash if there is no lease" do
        # actually also indirectly tested by many examples in this spec class.
        # Yet this explicit example can make things easier to debug if a problem arises.
        expect(subject.renew_exclusive_lease).to be_nil
      end

      it "renews the lease if there is one" do
        subject.try_obtain_exclusive_lease do
          expect(subject.renew_exclusive_lease).to be true
        end
      end
    end

    describe ".project_migrating?" do
      it "returns false if the project is not undergoing migration" do
        expect(described_class.project_migrating?(project)).to be false
      end

      it "returns true while the project is undergoing migration" do
        subject.enter do
          expect(described_class.project_migrating?(project)).to be true
        end
        expect(described_class.project_migrating?(project)).to be false
      end
    end
  end

  context 'Issues' do
    let!(:issue) do
      create(:issue,
        project: project,
        description: "Sample description with Git commit: #{git_sha}",
        title: "Sample title for Git commit: #{git_sha}")
    end

    it 'migrate title and description' do
      original_issue_timestamp = issue.updated_at
      subject.migrate_issues
      issue.reload
      expect(issue.description).to eq("Sample description with Git commit: #{hg_sha}")
      expect(issue.title).to eq("Sample title for Git commit: #{hg_sha}")
      expect_timestamp_unchanged(issue.updated_at, original_issue_timestamp)
    end

    context 'Notes' do
      let!(:my_comment) do
        create(:note_on_issue, project: project, noteable: issue,
          note: "Sample comment with Git commit: #{git_sha}")
      end

      it 'migrate comments' do
        original_issue_timestamp = issue.updated_at
        original_comment_timestamp = my_comment.updated_at
        subject.migrate_notes
        my_comment.reload
        expect(my_comment.note).to eq("Sample comment with Git commit: #{hg_sha}")
        expect_timestamp_unchanged(issue.updated_at, original_issue_timestamp)
        expect_timestamp_unchanged(my_comment.updated_at, original_comment_timestamp)
      end
    end
  end

  context 'Merge Requests' do
    let!(:merge_request) do
      create(:merge_request,
        source_project: project,
        source_branch: 'topic/default/feature',
        target_project: project,
        target_branch: 'branch/default',
        description: "Sample description with Git commit: #{git_sha}",
        title: "Sample title for Git commit: #{git_sha}")
    end

    it 'migrate title, description, and merge_request_diff_commits' do
      mr_diff = merge_request.merge_request_diffs.first
      original_mr_timestamp = merge_request.updated_at
      original_mr_diff_timestamp = mr_diff.updated_at
      expect(merge_request.all_commits.first.sha).to eq(git_sha)

      # migrate merge request
      subject.migrate_merge_requests
      merge_request.reload
      mr_diff.reload

      expect(merge_request.all_commits.first.sha).to eq(hg_sha)
      expect(merge_request.description).to eq("Sample description with Git commit: #{hg_sha}")
      expect(merge_request.title).to eq("Sample title for Git commit: #{hg_sha}")
      expect_timestamp_unchanged(merge_request.updated_at, original_mr_timestamp)
      expect_timestamp_unchanged(mr_diff.updated_at, original_mr_diff_timestamp)
    end

    context 'Notes' do
      let!(:my_comment) do
        create(:note_on_merge_request, project: project, noteable: merge_request,
          note: "Sample comment with Git commit: #{git_sha}")
      end

      it 'migrate comments' do
        original_mr_timestamp = merge_request.updated_at
        original_comment_timestamp = my_comment.updated_at
        subject.migrate_notes
        my_comment.reload
        expect(my_comment.note).to eq("Sample comment with Git commit: #{hg_sha}")
        expect_timestamp_unchanged(merge_request.updated_at, original_mr_timestamp)
        expect_timestamp_unchanged(my_comment.updated_at, original_comment_timestamp)
      end
    end

    context 'DiffNote' do
      let(:path) { "file2"}
      let!(:position) do
        Gitlab::Diff::Position.new(
          old_path: path,
          new_path: path,
          old_line: nil,
          new_line: 1,
          diff_refs: merge_request.diff_refs
        )
      end

      shared_examples 'DiffNote migration' do
        let!(:note) do
          create(:diff_note_on_merge_request,
                 project: project,
                 position: position,
                 noteable: merge_request,
                 note: "This reminds me of #{git_base_sha}")
        end

        it "migrates the note" do
          original_mr_timestamp = merge_request.updated_at
          original_note_timestamp = note.updated_at
          line_code_before = note.line_code
          subject.migrate_notes
          note.reload
          expect(note.note).to eq("This reminds me of #{hg_base_sha}")
          expect(note.line_code).to eq(line_code_before)
          expect_timestamp_unchanged(merge_request.updated_at, original_mr_timestamp)
          expect_timestamp_unchanged(note.updated_at, original_note_timestamp)
        end
      end

      context "when rugged Git implementation is enabled", :enable_rugged do
        it_behaves_like 'DiffNote migration'
      end

      context "when rugged Git implementation is not enabled" do
        it_behaves_like 'DiffNote migration'
      end
    end
  end

  context 'Pipelines' do
    let!(:pipeline) do
      create(:ci_pipeline,
        project: project,
        sha: git_sha,
        ref: 'topic/default/feature',
        status: 'success')
    end

    it 'migrate pipelines' do
      original_pipeline_timestamp = pipeline.updated_at
      subject.migrate_pipelines
      pipeline.reload
      expect(pipeline.sha).to eq(hg_sha)
      expect_timestamp_unchanged(pipeline.updated_at, original_pipeline_timestamp)
    end
  end

  context 'Releases' do
    let!(:release) do
      create(:release,
        project: project,
        description: "Sample description with Git commit: #{git_sha}")
    end

    it 'migrate description' do
      original_release_timestamp = release.updated_at
      subject.migrate_releases
      release.reload
      expect(release.description).to eq("Sample description with Git commit: #{hg_sha}")
      expect_timestamp_unchanged(release.updated_at, original_release_timestamp)
    end
  end

  describe '#vcs_type_enforcer' do
    let!(:project) { create(:native_hg_project, :hg_repository) }

    subject { described_class.new(project, project.creator) }

    it 'force repository to be HgGitRepository' do
      original_project_timestamp = project.updated_at
      expect(project.vcs_type).to eq("hg")
      subject.vcs_type_enforcer do
        expect(project.vcs_type).to eq("hg_git")
      end
      expect(project.vcs_type).to eq("hg")
      expect_timestamp_unchanged(project.updated_at, original_project_timestamp)
    end
  end

  describe '#do_migrate' do
    it "sets VCS Type to native even if there are errors if Merge Requests are migrated" do
      error = StandardError.new("oops")
      expect(subject).to receive(:migrate_issues).and_raise(error)
      expect { subject.do_migrate }.to raise_error(error)
      expect(project.vcs_type).to eq('hg')
    end

    it "does not set VCS Type to native if Merge Requests are not migrated" do
      error = StandardError.new("oops")
      expect(subject).to receive(:migrate_merge_requests).and_raise(error)
      expect { subject.do_migrate }.to raise_error(error)
      expect(project.vcs_type).to eq('hg_git')
    end

    context "about Notes" do
      let!(:issue) do
        create(:issue,
               project: project,
               description: "Sample description",
               title: "Sample title")
      end

      let!(:my_comment) do
        create(:note_on_issue, project: project, noteable: issue,
          note: "Sample comment")
      end

      it "returns errors" do
        expect(subject).to receive(:migrate_note_positions)
                             .and_raise(StandardError.new('failed'))
        expect(subject.migrate!).to eq({ notes: [my_comment.id] })
      end
    end
  end

  # Best approximation to equality of timestamps
  #
  # True equality of timestamps is probably impossible, because
  # truncation of nanoseconds occurs for reasons (would guess database
  # round-trips, but it does not matter).
  # Fortunately none of these examples can run in less than a microsecond
  def expect_timestamp_unchanged(ts1, ts2)
    expect((ts2 - ts1).abs).to be <= 1e-6
  end
end
